@@ -34,6 +34,7 @@
 #include "gtkmain.h"
 #include "gtksignal.h"
 #include "gtkmarshalers.h"
+#include "gtktreeprivate.h"
 
 #include <string.h>
 
@@ -108,7 +109,12 @@ static gboolean gtk_entry_completion_action_button_press (GtkWidget
                                                           gpointer                 user_data);
 static void     gtk_entry_completion_selection_changed   (GtkTreeSelection        *selection,
                                                           gpointer                 data);
-
+static gboolean	gtk_entry_completion_list_enter_notify	 (GtkWidget               *widget,
+							  GdkEventCrossing        *event,
+							  gpointer                 data);
+static gboolean gtk_entry_completion_list_motion_notify	 (GtkWidget		  *widget,
+							  GdkEventMotion          *event,
+							  gpointer 		   data);
 static void     gtk_entry_completion_insert_action       (GtkEntryCompletion      *completion,
                                                           gint                     index,
                                                           const gchar             *string,
@@ -270,7 +276,14 @@ gtk_entry_completion_init (GtkEntryCompletion *completion)
   g_signal_connect (priv->tree_view, "button_press_event",
                     G_CALLBACK (gtk_entry_completion_list_button_press),
                     completion);
+  g_signal_connect (priv->tree_view, "enter_notify_event",
+		    G_CALLBACK (gtk_entry_completion_list_enter_notify),
+		    completion);
+  g_signal_connect (priv->tree_view, "motion_notify_event",
+		    G_CALLBACK (gtk_entry_completion_list_motion_notify),
+		    completion);
   gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (priv->tree_view), FALSE);
+  _gtk_tree_view_set_hover_selection (GTK_TREE_VIEW (priv->tree_view), TRUE);
 
   sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->tree_view));
   gtk_tree_selection_set_mode (sel, GTK_SELECTION_SINGLE);
@@ -301,7 +314,14 @@ gtk_entry_completion_init (GtkEntryCompletion *completion)
   g_signal_connect (priv->action_view, "button_press_event",
                     G_CALLBACK (gtk_entry_completion_action_button_press),
                     completion);
+  g_signal_connect (priv->action_view, "enter_notify_event",
+		    G_CALLBACK (gtk_entry_completion_list_enter_notify),
+		    completion);
+  g_signal_connect (priv->action_view, "motion_notify_event",
+		    G_CALLBACK (gtk_entry_completion_list_motion_notify),
+		    completion);
   gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (priv->action_view), FALSE);
+  _gtk_tree_view_set_hover_selection (GTK_TREE_VIEW (priv->action_view), TRUE);
 
   sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->action_view));
   gtk_tree_selection_set_mode (sel, GTK_SELECTION_SINGLE);
@@ -790,6 +810,7 @@ gtk_entry_completion_get_entry (GtkEntryCompletion *completion)
  *
  * Sets the model for a #GtkEntryCompletion. If @completion already has
  * a model set, it will remove it before setting the new model.
+ * If @model is %NULL, then it will unset the model.
  *
  * Since: 2.4
  */
@@ -798,7 +819,7 @@ gtk_entry_completion_set_model (GtkEntryCompletion *completion,
                                 GtkTreeModel       *model)
 {
   g_return_if_fail (GTK_IS_ENTRY_COMPLETION (completion));
-  g_return_if_fail (GTK_IS_TREE_MODEL (model));
+  g_return_if_fail (model == NULL || GTK_IS_TREE_MODEL (model));
 
   /* code will unref the old filter model (if any) */
   completion->priv->filter_model =
@@ -960,10 +981,10 @@ gtk_entry_completion_insert_action (GtkEntryCompletion *completion,
 /**
  * gtk_entry_completion_insert_action_text:
  * @completion: A #GtkEntryCompletion.
- * @index: The index of the item to insert.
+ * @index_: The index of the item to insert.
  * @text: Text of the item to insert.
  *
- * Inserts an action in @completion's action item list at position @index
+ * Inserts an action in @completion's action item list at position @index_
  * with text @text. If you want the action item to have markup, use
  * gtk_entry_completion_insert_action_markup().
  *
@@ -971,57 +992,57 @@ gtk_entry_completion_insert_action (GtkEntryCompletion *completion,
  */
 void
 gtk_entry_completion_insert_action_text (GtkEntryCompletion *completion,
-                                         gint                index,
+                                         gint                index_,
                                          const gchar        *text)
 {
   g_return_if_fail (GTK_IS_ENTRY_COMPLETION (completion));
   g_return_if_fail (text != NULL);
 
-  gtk_entry_completion_insert_action (completion, index, text, FALSE);
+  gtk_entry_completion_insert_action (completion, index_, text, FALSE);
 }
 
 /**
  * gtk_entry_completion_insert_action_markup:
  * @completion: A #GtkEntryCompletion.
- * @index: The index of the item to insert.
+ * @index_: The index of the item to insert.
  * @markup: Markup of the item to insert.
  *
- * Inserts an action in @completion's action item list at position @index
+ * Inserts an action in @completion's action item list at position @index_
  * with markup @markup.
  *
  * Since: 2.4
  */
 void
 gtk_entry_completion_insert_action_markup (GtkEntryCompletion *completion,
-                                           gint                index,
+                                           gint                index_,
                                            const gchar        *markup)
 {
   g_return_if_fail (GTK_IS_ENTRY_COMPLETION (completion));
   g_return_if_fail (markup != NULL);
 
-  gtk_entry_completion_insert_action (completion, index, markup, TRUE);
+  gtk_entry_completion_insert_action (completion, index_, markup, TRUE);
 }
 
 /**
  * gtk_entry_completion_delete_action:
  * @completion: A #GtkEntryCompletion.
- * @index: The index of the item to Delete.
+ * @index_: The index of the item to Delete.
  *
- * Deletes the action at @index from @completion's action list.
+ * Deletes the action at @index_ from @completion's action list.
  *
  * Since: 2.4
  */
 void
 gtk_entry_completion_delete_action (GtkEntryCompletion *completion,
-                                    gint                index)
+                                    gint                index_)
 {
   GtkTreeIter iter;
 
   g_return_if_fail (GTK_IS_ENTRY_COMPLETION (completion));
-  g_return_if_fail (index >= 0);
+  g_return_if_fail (index_ >= 0);
 
   gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (completion->priv->actions),
-                                 &iter, NULL, index);
+                                 &iter, NULL, index_);
   gtk_list_store_remove (completion->priv->actions, &iter);
 }
 
@@ -1093,6 +1114,29 @@ get_borders (GtkEntry *entry,
     }
 }
 
+static gboolean
+gtk_entry_completion_list_enter_notify (GtkWidget        *widget,
+					GdkEventCrossing *event,
+					gpointer          data)
+{
+  GtkEntryCompletion *completion = GTK_ENTRY_COMPLETION (data);
+  
+  return completion->priv->ignore_enter;
+}
+
+static gboolean
+gtk_entry_completion_list_motion_notify (GtkWidget      *widget,
+					 GdkEventMotion *event,
+					 gpointer        data)
+{
+  GtkEntryCompletion *completion = GTK_ENTRY_COMPLETION (data);
+
+  completion->priv->ignore_enter = FALSE; 
+  
+  return FALSE;
+}
+
+
 /* some nasty size requisition */
 gboolean
 _gtk_entry_completion_resize_popup (GtkEntryCompletion *completion)
@@ -1103,9 +1147,11 @@ _gtk_entry_completion_resize_popup (GtkEntryCompletion *completion)
   gint monitor_num;
   GdkRectangle monitor;
   GtkRequisition popup_req;
+  GtkRequisition entry_req;
   GtkTreePath *path;
   gboolean above;
-
+  gint width;
+  
   gdk_window_get_origin (completion->priv->entry->window, &x, &y);
   get_borders (GTK_ENTRY (completion->priv->entry), &x_border, &y_border);
 
@@ -1124,9 +1170,14 @@ _gtk_entry_completion_resize_popup (GtkEntryCompletion *completion)
   else
     gtk_widget_show (completion->priv->scrolled_window);
 
+  screen = gtk_widget_get_screen (GTK_WIDGET (completion->priv->entry));
+  monitor_num = gdk_screen_get_monitor_at_window (screen, 
+						  GTK_WIDGET (completion->priv->entry)->window);
+  gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
+
+  width = MIN (completion->priv->entry->allocation.width, monitor.width);
   gtk_widget_set_size_request (completion->priv->tree_view,
-                               completion->priv->entry->allocation.width - 2 * x_border,
-                               items * height);
+                               width - 2 * x_border, items * height);
 
   /* default on no match */
   completion->priv->current_selected = -1;
@@ -1142,27 +1193,22 @@ _gtk_entry_completion_resize_popup (GtkEntryCompletion *completion)
                                           &height);
 
       gtk_widget_set_size_request (completion->priv->action_view,
-                                   completion->priv->entry->allocation.width - 2 * x_border,
-                                   items * height);
+                                   width - 2 * x_border, items * height);
     }
   else
     gtk_widget_hide (completion->priv->action_view);
 
   gtk_widget_size_request (completion->priv->popup_window, &popup_req);
-
-  screen = gtk_widget_get_screen (GTK_WIDGET (completion->priv->entry));
-  monitor_num = gdk_screen_get_monitor_at_window (screen, 
-						  GTK_WIDGET (completion->priv->entry)->window);
-  gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
+  gtk_widget_size_request (completion->priv->entry, &entry_req);
   
   if (x < monitor.x)
     x = monitor.x;
   else if (x + popup_req.width > monitor.x + monitor.width)
     x = monitor.x + monitor.width - popup_req.width;
   
-  if (y + height + popup_req.height <= monitor.y + monitor.height)
+  if (y + entry_req.height + popup_req.height <= monitor.y + monitor.height)
     {
-      y += height;
+      y += entry_req.height;
       above = FALSE;
     }
   else
@@ -1193,8 +1239,8 @@ _gtk_entry_completion_popup (GtkEntryCompletion *completion)
   if (GTK_WIDGET_MAPPED (completion->priv->popup_window))
     return;
 
-  completion->priv->may_wrap = TRUE;
-
+  completion->priv->ignore_enter = TRUE;
+    
   column = gtk_tree_view_get_column (GTK_TREE_VIEW (completion->priv->action_view), 0);
   renderers = gtk_tree_view_column_get_cell_renderers (column);
   gtk_widget_ensure_style (completion->priv->tree_view);
@@ -1220,6 +1266,11 @@ _gtk_entry_completion_popup (GtkEntryCompletion *completion)
 void
 _gtk_entry_completion_popdown (GtkEntryCompletion *completion)
 {
+  if (!GTK_WIDGET_MAPPED (completion->priv->popup_window))
+    return;
+
+  completion->priv->ignore_enter = FALSE;
+  
   gdk_pointer_ungrab (GDK_CURRENT_TIME);
   gtk_grab_remove (completion->priv->popup_window);
 
